import glob
import subprocess as sp


def check_blastdb(db_path):
    file_extensions = ['nhr', 'nin', 'nsq']
    for fe in file_extensions:
        files = glob.glob(f'{db_path}*{fe}')
        if not files:
            # At least one of the BLAST db files is not existent
            return False
    return True

def make_blastndb(inpath, outpath):
    db_command = 'makeblastdb -in {} -out {} -dbtype nucl'.format(inpath, outpath)
    sp.call(db_command, shell=True)



"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""

import glob
import pyfaidx
import subprocess as sp
import sys
import os

from utils import check_blastdb
from utils import make_blastndb


def cmsearcher(mirna, cm_cutoff, cpu, msl, models, query, blastdb, out, cleanup, heuristic):
    """
    Parameters
    ----------
    mirna       :   Mirna object
    cm_cutoff   :   Minimum bit score of the CMsearch hit relative to the
                    bit score that results from searching in the reference genome with the same model
    cpu         :   Number of cores to use
    msl             Minimum length of CMsearch hit relative to the reference pre-miRNA length
    models      :   Path to the directory that contains the CMs
    query       :   Path to the query genome
    qblast      :   Optional Path to a BLASTdb of the query genome
    out         :   Path to the output directory
    cleanup     :   Delete intermediate files True/False
    heuristic   :   Should heuristic mode be used and if yes, set parameters (True/False, evalue, minlength)

    Returns
    -------
    cm_results : Dictionary containing hits of the cmsearch

    """
    mirna_id = mirna.name

    # blastdb = query.replace('.fa', '')
    # Calculate the bit score cutoff.
    cut_off = mirna.bit * cm_cutoff
    # Calculate the length cutoff.
    len_cut = len(mirna.pre) * msl
    blast_len_cut = len(mirna.pre) * heuristic[2]
    # if no model exists for mirna return False
    if not os.path.isfile('{}/{}.cm'.format(models, mirna_id)):
        print('# No model found for {}. Skipping..'.format(mirna_id))
        return False
    # Perform covariance model search.
    # Report and inclusion thresholds set according to cutoff.

    if heuristic[0]:
        # extract candidate regions
        print('# Identifying candidate regions for cmsearch heuristic')
        tmp_out = '{0}/tmp_blast_{1}.out'.format(out, mirna_id)
        with open(tmp_out, 'w') as inf:
            inf.write(">{}\n{}\n".format(mirna_id, mirna.pre))
        blast_command = (
            'blastn -evalue {3} -task blastn -db {0} -query {1} '
            '-num_threads {2} -outfmt "6 sseqid sstart send sstrand length"'.format(blastdb, tmp_out, cpu, heuristic[1])
        )
        # print(blast_command)
        res = sp.run(blast_command, shell=True, capture_output=True)
        os.remove(tmp_out)
        if res.stdout:
            print('Blast step finished')
        else:
            cm_results = False
            return cm_results
        result = res.stdout.decode('utf-8').split('\n')
        hit_list = [line.split() for line in result if line and float(line.split()[-1]) >= blast_len_cut]
        # hit_list = [line.split() for line in result if line]
        print(f'# Found {len(hit_list)} BLAST hits of the reference '
              f'pre-miRNA in the query')
        genes = pyfaidx.Fasta(query)
        # collect heuristic sequences
        print('collecting sequences')
        # set border regions to include in analysis
        extraregion = 1000
        diff_dict = {}
        hit_at_start = False
        hit_at_end = False
        heuristic_fa = '{0}/heuristic_{1}.out'.format(out, mirna_id)
        with open(heuristic_fa, 'w') as of:
            for hit in hit_list:
                chrom, start, end, strand, length = hit
                strand = strand.replace('plus', '+')
                strand = strand.replace('minus', '-')
                if strand == '-':
                    start, end = end, start
                header = ">{}|{}|{}|{}\n".format(chrom, start, end, strand)
                # print(header)
                start = int(start)
                end = int(end)
                if start > extraregion:
                    n_start = start - extraregion
                else:
                    # hit starts at the beginning of the chromosome
                    n_start = 0
                    hit_at_start = True
                n_end = end + extraregion
                if n_end > genes[chrom][-1].end:
                    # hit is at the end of the chromosome
                    n_end = genes[chrom][-1].end
                    hit_at_end = True
                if strand == '+':
                    sequence = genes[chrom][n_start:n_end].seq
                elif strand == '-':
                    sequence = genes[chrom][n_start:n_end].reverse.complement.seq
                header = ">{}|{}|{}|{}|{}|{}\n".format(chrom, start, end, strand, hit_at_start, hit_at_end)
                of.write(header)
                of.write(f'{sequence}\n')
        # start the cmsearch with the heuristic candidates
        heuristic_cms = '{0}/cmsearch_heuristic.out'.format(out, mirna_id)
        # if not os.path.isfile(cms_output):
        print('# Running covariance model search for {}'.format(mirna_id))
        cms_command = (
            'cmsearch -T {5} --incT {5} --cpu {0} --noali '
            '--tblout {1} {2}/{3}.cm {4}'
                .format(cpu, heuristic_cms, models, mirna_id, heuristic_fa, cut_off)
        )
        res = sp.run(cms_command, shell=True, capture_output=True)
        if not res.stdout:
            print('# cmsearch did not find anything')
            cm_results = False
            return cm_results

        # else:
        #     print('# Found cm_search results at: {}. Using those'.format(cms_output))
        #     cm_results = cmsearch_parser(cms_output, cut_off, len_cut, mirna_id)
        heur_results = heuristic_cms.replace('.out', '_res.out')
        with open(heuristic_cms, 'r') as inf, open(heur_results, 'w') as of:
            for line in inf:
                if not line.startswith('#'):
                    data = line.strip().split()
                    blast_chrom = data[0].split('|')[0]
                    blast_start, blast_end = map(int, data[0].split('|')[1:3])
                    hit_at_start, hit_at_end = data[0].split('|')[4:6]
                    blast_strand = data[0].split('|')[3]
                    cm_start, cm_end = map(int, data[7:9])
                    # cm_start, cm_end = map(int, data[5:7])
                    cm_strand = data[9]
                    if cm_strand == '-':
                        # reverse hits should not be possible since blasthits were extracted
                        # as reverse complement if they were on the minus strand
                        continue
                    # if not cm_strand == blast_strand:
                    #     print('Rejecting cmsearch hit. Not on same strand as BLAST hit')
                    #     break
                    # print(type(hit_at_start), hit_at_end)
                    if hit_at_start == 'True':
                        print('# cmsearch hit for {} at the beginning of a chromosome in the query species'.format(mirna_id))
                        hit_start = cm_start
                        hit_end = cm_end
                    elif hit_at_end == 'True':
                        print('# cmsearch hit for {} at the end of a chromosome in the query species'.format(
                            mirna_id))
                        hit_start = blast_start + (cm_start - extraregion) - 1
                        hit_end = hit_start + (cm_end - cm_start) - 1
                    else:
                        hit_start = blast_start + (cm_start - extraregion) - 1
                        hit_end = hit_start + (cm_end - extraregion) - 1
                    # hit_start = blast_end - cm_end
                    # hit_end = blast_end - cm_start
                    data[0] = blast_chrom
                    data[7] = hit_start
                    data[8] = hit_end
                    data[9] = blast_strand
                    data = [str(entry) for entry in data]
                    of.write('\t'.join(data))
                    of.write('\n')
        cm_results = cmsearch_parser(heur_results, cut_off, len_cut, mirna_id)
        if cleanup:
            os.remove(heuristic_fa)
            os.remove(heur_results)
            os.remove(heuristic_cms)
    else:
        cms_output = '{0}/cmsearch_{1}.out'.format(out, mirna_id)
        if not os.path.isfile(cms_output):
            print('# # Running covariance model search for {}'.format(mirna_id))
            cms_command = (
                'cmsearch -T {5} --incT {5} --cpu {0} --noali '
                '--tblout {1} {2}/{3}.cm {4}'
                    .format(cpu, cms_output, models, mirna_id, query, cut_off)
            )
            sp.call(cms_command, shell=True)
        else:
            print('# Found cm_search results at: {}. Using those'.format(cms_output))
        cm_results = cmsearch_parser(cms_output, cut_off, len_cut, mirna_id)

    return cm_results


# cmsearch_parser: Parse the output of cmsearch while eliminating
#                  duplicates and filtering entries according to the
#                  defined cutoff.
def cmsearch_parser(cms, cmc, lc, mirid):
    """
    Parse the output of cmsearch while eliminating duplicates and filtering
    entries according to the defined cutoff.

    Parameters
    ----------
    cms     :   Path to cmsearch output
    cmc     :   Cutoff to decide which candidate hits should be included for the reverse BLAST search
    lc      :   Length cutoff
    mirid   :   miRNA ID

    Returns
    -------
    hits_dict : Dictionary containing hits of the cmsearch

    """
    # Output
    hits_dict = {}
    # Required for finding duplicates, stores hits per chromosome
    chromo_dict = {}
    cut_off = cmc

    with open(cms) as cmsfile:
        # Collect only the hits which satisfy the bit score cutoff.
        hits = [
            line.strip().split() for line in cmsfile
            if not line.startswith('#')
               and float(line.strip().split()[14]) >= cut_off
               and abs(int(line.split()[7]) - int(line.strip().split()[8])) >= lc
        ]
        # Add the hits to the return dictionary.
        if hits:
            for candidate_nr, hit in enumerate(hits, 1):
                h_chrom = hit[0]
                h_start = hit[7]
                h_end = hit[8]
                h_strand = hit[9]
                h_score = hit[14]
                # blastparser expects the start to be the smaller number, will extract reverse complement if on - strand
                if h_start > h_end:
                    start_tmp = h_start
                    h_start = h_end
                    h_end = start_tmp
                data = (
                    '{0}_c{1}'.format(mirid, candidate_nr),
                    h_chrom, h_start, h_end, h_strand, h_score
                )
                hits_dict[data[0]] = data

                # Store the hits that satisfy the bit score cutoff to filter
                # duplicates.
                try:
                    chromo_dict[data[1]].append(data)
                except:
                    chromo_dict[data[1]] = [data]

    # Loop over the candidate hits to eliminate duplicates.
    for chromo in chromo_dict:
        nrhits = len(chromo_dict[chromo])
        if nrhits > 1:
            for hitnr in range(nrhits):
                start = int(chromo_dict[chromo][hitnr][2])
                stop = int(chromo_dict[chromo][hitnr][3])
                strand = chromo_dict[chromo][hitnr][4]
                score = float(chromo_dict[chromo][hitnr][5])
                for chitnr in range(hitnr +1, nrhits):
                    if strand != chromo_dict[chromo][chitnr][4]:
                        cstart = int(chromo_dict[chromo][chitnr][2])
                        cstop = int(chromo_dict[chromo][chitnr][3])
                        cscore = float(chromo_dict[chromo][chitnr][5])
                        # Test if the two hits from opposite strands
                        # overlap, which means one of them is
                        # (probably) a false positive.
                        # Out of two conflicting hits, the one with the
                        # highest cmsearch bit score is retained.
                        if (
                                start in range(cstart, cstop +1)
                                or stop in range(cstart, cstop +1)
                                or cstart in range(start, stop +1)
                                or cstop in range(start, stop +1)
                        ):
                            if score > cscore:
                                try:
                                    del hits_dict[chromo_dict[chromo]
                                    [chitnr][0]]
                                except:
                                    pass
                            else:
                                try:
                                    del hits_dict[chromo_dict[chromo]
                                    [hitnr][0]]
                                except:
                                    pass
    return hits_dict



import sys
import re

# Parse an Ensembl GTF file to store the coordinates for each protein-coding gene in a
# dictionary
def gtf_parser(gtf):
    species_name = gtf.split('/')[-1].split('.')[0]
    chr_dict = {}
    chromo = ''

    #with open(inpath) as infile, open(outpath, 'wb') as outfile:
    with open(gtf) as infile:
        for line in infile:
            if (
                not line.startswith('#')
                and line.split()[2] == 'gene'
                and line.split('gene_biotype')[1].split('\"')[1]
                == 'protein_coding'
            ):
                linedata = line.strip().split('\t')
                contig = linedata[0]
                geneid = linedata[-1].split('\"')[1]
                start = int(linedata[3])
                end = int(linedata[4])
                strand = linedata[6]
                if contig != chromo:
                    i = 1
                    chromo = contig
                chr_dict[geneid] = (contig, i)
                try:
                    chr_dict[contig][i] = (geneid, start, end, strand)
                except:
                    chr_dict[contig] = {i: (geneid, start, end, strand)}
                i += 1
    return chr_dict


def gff_parser(gff, id_type):
    """

    Parameters
    ----------
    gff :   Path to a RefSeq gff3 file
    id  :   type of identifier in the gff3 file that is to be compared against the IDs in the pairwise ortholgs file

    Returns
    -------
    Dictionary containing the location of each gene on a chromosome
    {'gene_id': ('chromosome': position), 'chromosome': {position: ('gene_id', start, end, strand)} }

    """
    chr_dict = {}
    chromo = ''
    # with open(inpath) as infile, open(outpath, 'wb') as outfile:
    with open(gff) as infile:
        for line in infile:
            if (
                    not line.startswith('#')
                    and line.split('\t')[2] == 'gene'
                    and 'gene_biotype=protein_coding' in line.split()[-1]
            ):
                linedata = line.strip().split('\t')
                if id_type in ['ID', 'Name', 'gene_id']:
                    geneid = linedata[-1].split(f'{id_type}=')[1].split(';')[0]
                elif id_type == 'GeneID':
                    geneid = re.split('[;,]', linedata[-1].split(f'{id_type}:')[1])[0]
                else:
                    print('ERROR: Unknown identifier type "{}"'.format(id_type))
                    sys.exit()

                contig = linedata[0]
                start = int(linedata[3])
                end = int(linedata[4])
                strand = linedata[6]
                if contig != chromo:
                    i = 1
                    chromo = contig
                chr_dict[geneid] = (contig, i)
                try:
                    chr_dict[contig][i] = (geneid, start, end, strand)
                except:
                    chr_dict[contig] = {i: (geneid, start, end, strand)}
                i += 1
    return chr_dict

###############################################################################
#Try to find the ortholog for a given reference gene in a core set species
def ortho_search(r_gene, ortho_dict):
    orthologs = {}
    for core_taxon in ortho_dict.keys():
        try:
            ortholog = ortho_dict[core_taxon][r_gene]
            orthologs[core_taxon] = ortholog
            print(
                '{0} is the ortholog for {1} in {2}.'
                .format(ortholog, r_gene, core_taxon)
            )
        except:
            print(
                'No ortholog found for {0} in {1}.'
                .format(r_gene, core_taxon)
            )
    return orthologs



"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="ncOrtho",
    version="0.0.1",
    python_requires='>=3.7.0',
    description=" Targeted ortholog search for miRNAs ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Felix Langschied",
    author_email="langschied@bio.uni-frankfurt.de",
    url="https://github.com/felixlangschied/ncortho",
    packages=find_packages(),
    package_data={'': ['*']},
    install_requires=[
        'PyYAML',
        'biopython',
        'pyfaidx'
    ],
    entry_points={
        'console_scripts': ["ncCreate = ncOrtho.coreset.coreset:main",
                            "ncSearch = ncOrtho.ncortho:main"
                            ],
    },
    license="GPL-3.0",
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
)



"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""


# Modules import
import argparse
import multiprocessing as mp
import os
import subprocess as sp
import sys

# Internal ncOrtho modules
from blastparser import BlastParser
from blastparser import ReBlastParser
from genparser import GenomeParser
from cmsearch import cmsearcher
from utils import check_blastdb
from utils import make_blastndb

###############################################################################


# Central class of microRNA objects
class Mirna(object):
    def __init__(self, name, chromosome, start, end, strand, pre, mature, bit):
        # miRNA identifier
        self.name = name
        # chromosome that the miRNA is located on
        if chromosome.startswith('chr'):
            self.chromosome = chromosome.split('chr')[1]
        else:
            self.chromosome = chromosome
        # start position of the pre-miRNA
        self.start = int(start)
        # end position of the pre-miRNA
        self.end = int(end)
        # sense (+) or anti-sense (-) strand
        self.strand = strand
        # nucleotide sequence of the pre-miRNA
        self.pre = pre.replace('U', 'T')
        # nucleotide sequence of the mature miRNA
        self.mature = mature.replace('U', 'T')
        # reference bit score that miRNA receives by its own
        # covariance model
        self.bit = bit


def mirna_maker(mirpath, cmpath, output, msl):
    """
    Parses the miRNA data input file and returns a dictionary of Mirna objects.

    Parameters
    ----------
    mirpath : STR
        Path to file with microRNA data.
    cmpath : STR
        Path to covariance models.
    output : STR
        Path for writing temporary files.
    msl : FLOAT
        Length filter.

    Returns
    -------
    mmdict : DICT
        Dictionary with Mirna objects.

    """
    
    mmdict = {} # will be the return object
    
    with open(mirpath) as mirna_file:
        mirna_data = [
            line.strip().split() for line in mirna_file
            if not line.startswith('#')
        ]

    for mirna in mirna_data:
        mirid = mirna[0]
        seq = mirna[5]
        query = '{0}/{1}.fa'.format(output, mirid)
        model = '{0}/{1}.cm'.format(cmpath, mirid)
        # Check if the covariance model even exists, otherwise skip to
        # the next miRNA.
        if not os.path.isfile(model):
            print('# No covariance model found for {}'.format(mirid))
            continue
        # Check if the output folder exists, otherwise create it.
        if not os.path.isdir('{}'.format(output)):
            mkdir = 'mkdir -p {}'.format(output)
            sp.call(mkdir, shell=True)


        # Obtain the reference bit score for each miRNA by applying it
        # to its own covariance model.
        print('# Calculating reference bit score for {}.'.format(mirid))
        
        # Create a temporary FASTA file with the miRNA sequence as
        # query for external search tool cmsearch to calculate the
        # reference bit score.
        with open(query, 'w') as tmpfile:
            tmpfile.write('>{0}\n{1}'.format(mirid, seq))
        cms_output = '{0}/ref_cmsearch_{1}_tmp.out'.format(output, mirid)
        cms_log = '{0}/ref_cmsearch_{1}.log'.format(output, mirid)
        if not os.path.isfile(cms_output):
            cms_command = (
                'cmsearch -E 0.01 --noali -o {3} --tblout {0} {1} {2}'
                .format(cms_output, model, query, cms_log)
            )
            sp.call(cms_command, shell=True)
        else:
            print('# Found cmsearch results at: {} using those'.format(cms_output))
        with open(cms_output) as cmsfile:
            hits = [
                line.strip().split() for line in cmsfile
                if not line.startswith('#')
            ]
            if hits:
                top_score = float(hits[0][14])
            # In case of any issues occuring in the calculation of the bit
            # score, no specific threshold can be determined. The value will
            # set to zero, which technically turns off the filter.
            else:
                print(
                    '# Warning: Self bit score not applicable, '
                    'setting threshold to 0.'
                )
                top_score = 0.0

        mirna.append(top_score)

        # Remove temporary files.
        for rmv_file in [cms_output, cms_log, query]:
            sp.call('rm {}'.format(rmv_file), shell=True)

        # Create output.
        mmdict[mirna[0]] = Mirna(*mirna)

    return mmdict


# Write a FASTA file containing the accepted orthologs.
def write_output(a, o, cm):
    """
    Parameters
    ----------
    a   :   Dictionary of accepted hits
    o   :   Output path
    dm  :   Dictionary of cmsearch output

    Returns
    -------
    None.

    """
    with open(o, 'w') as outfile:
        for hit in a:
            tup = cm[hit]
            header = '|'.join(list(tup))
            outfile.write('>{0}\n{1}\n'.format(header, a[hit]))


# Allow boolean argument parsing
def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


# Main function
def main():

    # Print header
    print('\n'+'#'*57)
    print('###'+' '*51+'###')
    print('###   ncOrtho - ortholog search for non-coding RNAs   ###')
    print('###'+' '*51+'###')
    print('#'*57+'\n')
    
    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description='Find orthologs of reference miRNAs in the genome of a query species.'
    )
    # covariance models folder
    parser.add_argument(
        '-m', '--models', metavar='<path>', type=str,
        help='Path to directory containing covariance models (.cm)'
    )
    # mirna data
    parser.add_argument(
        '-n', '--ncrna', metavar='<path>', type=str,
        help='Path to Tab separated file with information about the reference miRNAs'
    )
    # output folder
    parser.add_argument(
        '-o', '--output', metavar='<path>', type=str,
        help='Path to the output directory'
    )
    # query genome
    parser.add_argument(
        '-q', '--query', metavar='<.fa>', type=str,
        help='Path to query genome in FASTA format'
    )
    # reference genome
    parser.add_argument(
        '-r', '--reference', metavar='<.fa>', type=str,
        help='Path to reference genome in FASTA format'
    )
    ##########################################################################
    # Optional Arguments
    ##########################################################################
    # query_name
    parser.add_argument(
        '--queryname', metavar='str', type=str, nargs='?', const='', default='',
        help=(
            'Name for the output directory '
            '(Use this if you run ncOrtho for multiple taxa whose genome file has the same name)'
        )
    )
    # cpu, use maximum number of available cpus unless specified otherwise
    parser.add_argument(
        '--cpu', metavar='int', type=int,
        help='number of cpu cores ncOrtho should use', nargs='?',
        const=mp.cpu_count(), default=mp.cpu_count()
    )
    # bit score cutoff for cmsearch hits
    parser.add_argument(
        '--cm_cutoff', metavar='float', type=float,
        help='cmsearch bit score cutoff', nargs='?', const=0.5, default=0.5
    )
    # length filter to prevent short hits
    parser.add_argument(
        '--minlength', metavar='float', type=float,
        help='CMsearch hit in the query species must have at '
             'least the length of this value times the length of the refernce pre-miRNA',
        nargs='?', const=0.7, default=0.7
    )
    parser.add_argument(
        '--heuristic', type=str2bool, metavar='True/False', nargs='?', const=True, default=True,
        help=(
            'Perform a BLAST search of the reference miRNA in the query genome to identify '
            'candidate regions for the CMsearch. Majorly improves speed. (Default: True)'
        )
    )
    parser.add_argument(
        '--heur_blast_evalue', type=float, metavar='float', nargs='?', const=0.5, default=0.5,
        help=(
            'Evalue filter for the BLASTn search that '
            'determines candidate regions for the CMsearch when running ncOrtho in heuristic mode. '
            '(Default: 0.5) (Set to 10 to turn off)'
        )
    )
    parser.add_argument(
        '--heur_blast_length', type=float, metavar='float', nargs='?', const=0.5, default=0.5,
        help=(
            'Length cutoff relative to the refernce pre-miRNA length for the BLASTn search that '
            'determines candidate regions for the CMsearch when running ncOrtho in heuristic mode. '
            '(Default: 0.5) (Set to 0 to turn off)'
        )
    )
    parser.add_argument(
        '--cleanup', type=str2bool, metavar='True/False', nargs='?', const=True, default=True,
        help=(
            'Cleanup temporary files (Default: True)'
        )
    )
    parser.add_argument(
        '--refblast', type=str, metavar='<path>', nargs='?', const='', default='',
        help=(
            'Path to BLASTdb of the reference species'
        )
    )
    parser.add_argument(
        '--queryblast', type=str, metavar='<path>', nargs='?', const='', default='',
        help=(
            'Path to BLASTdb of the query species'
        )
    )
    parser.add_argument(
        '--maxcmhits', metavar='int', nargs='?', const=50, default=50,
        help=(
            'Maximum number of cmsearch hits to examine. '
            'Decreases runtime significantly if reference miRNA in genomic repeat region (Default: 50). '
            'Set to empty variable to disable (i.e. --maxcmhits="")'
        )
    )
    # use dust filter?
    parser.add_argument(
        '--dust', metavar='yes/no', type=str,
        help='Use BLASTn dust filter during re-BLAST. Greatly decreases runtime if reference miRNA(s) '
             'are located in repeat regions. '
             'However ncOrtho will also not identify orthologs for these miRNAs',
        nargs='?',
        const='no', default='no'
    )
    # check Co-ortholog-ref
    parser.add_argument(
        '--checkCoorthologsRef', type=str2bool, metavar='True/False', nargs='?', const=False, default=False,
        help=(
            'If the re-blast does not identify the original reference miRNA sequence as best hit,'
            'ncOrtho will check whether the best blast '
            'hit is likely a co-ortholog of the reference miRNA relative to the search taxon. '
            'NOTE: Setting this flag will substantially increase'
            'the sensitivity of HaMStR but most likely affect also the specificity, '
            'especially when the search taxon is evolutionarily only very'
            'distantly related to the reference taxon (Default: False)'
        )
    )

    # Show help when no arguments are added.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()
    
    # Check if computer provides the desired number of cores.
    available_cpu = mp.cpu_count()
    if args.cpu > available_cpu:
        print(
            '# Error: The provided number of CPU cores is higher than the '
            'number available on this system. Exiting...'
        )
        sys.exit(1)
    else:
        cpu = args.cpu

    mirnas = args.ncrna
    models = args.models
    output = args.output
    query = args.query
    reference = args.reference

    # optional
    max_hits = args.maxcmhits
    cm_cutoff = args.cm_cutoff
    checkCoorthref = args.checkCoorthologsRef
    cleanup = args.cleanup
    heuristic = (args.heuristic, args.heur_blast_evalue, args.heur_blast_length)
    msl = args.minlength
    refblast = args.refblast
    qblast = args.queryblast
    dust = args.dust.strip()

    ##########################################################################################
    # Starting argument checks
    ##########################################################################################

    # Test if optional query name was given
    if args.queryname:
        qname = args.queryname
    else:
        qname = '.'.join(query.split('/')[-1].split('.')[:-1])

    # Test if input files exist
    all_files = [mirnas, reference, query]
    for pth in all_files:
        if not os.path.isfile(pth):
            print(f'ERROR: {pth} is not a file')
            sys.exit()
    if not os.path.isdir(models):
        print(f'ERROR: Directory with covariance models does not exist at: {models}')
        sys.exit()

    # create symbolic links to guarantee writing permissions
    q_data = '{}/data'.format(output)
    if not os.path.isdir(q_data):
        os.makedirs(q_data)
    qlink = f'{q_data}/{qname}.fa'
    try:
        os.symlink(query, qlink)
    except FileExistsError:
        pass

    # check if reference BLASTdb was given as input
    if refblast:
        # check if BLASTdb exists
        if not check_blastdb(refblast):
            print('# Reference BLASTdb not found at: {}'.format(refblast))
            sys.exit()
    else:
        refname = reference.split('/')[-1]
        refblast = f'{q_data}/{refname}'
        # check if BLASTdb exists already. Otherwise create it
        if not check_blastdb(refblast):
            make_blastndb(reference, refblast)
    # check if query BLASTdb was given as input (only used in heuristic mode)
    if qblast and heuristic:
        # check if qblast exists
        if not check_blastdb(qblast):
            print('# Query BLASTdb not found at: {}'.format(refblast))
            sys.exit()
    elif heuristic:
        print('# Creating query Database')
        qblast = qlink
        make_blastndb(query, qlink)


    ################################################################################################
    # Main
    ###############################################################################################

    # Create miRNA objects from the list of input miRNAs.
    mirna_dict = mirna_maker(mirnas, models, output, msl)
    # Print out query
    print('### Starting ncOrtho run for {}'.format(query))

    # Identify ortholog candidates.
    for mir_data in mirna_dict:
        sys.stdout.flush()
        mirna = mirna_dict[mir_data]
        mirna_id = mirna.name
        if heuristic:
            outdir = '{}'.format(output)
        else:
            outdir = '{}/{}'.format(output, mirna_id)
        # Create output folder, if not existent.
        if not os.path.isdir(outdir):
            os.mkdir(outdir)
        # start cmsearch
        cm_results = cmsearcher(mirna, cm_cutoff, cpu, msl, models, qlink, qblast, output, cleanup, heuristic)

        # Extract sequences for candidate hits (if any were found).
        if not cm_results:
            print('# No hits found for {}.\n'.format(mirna_id))
            continue
        elif max_hits and len(cm_results) > max_hits:
            print('# Maximum CMsearch hits reached. Restricting to best {} hits'.format(max_hits))
            cm_results = {k: cm_results[k] for k in list(cm_results.keys())[:max_hits]}
        # get sequences for each CM result
        gp = GenomeParser(qlink, cm_results.values())
        candidates = gp.extract_sequences()
        nr_candidates = len(candidates)
        if nr_candidates == 1:
            print(
                '\n# Covariance model search successful, found 1 '
                'ortholog candidate.\n'
            )
        else:
            print(
                '\n# Covariance model search successful, found {} '
                'ortholog candidates.\n'
                .format(nr_candidates)
            )
        print('# Evaluating candidates.\n')
        
        # Perform reverse BLAST test to verify candidates, stored in
        # a list (accepted_hits).
        reblast_hits = {}
        for candidate in candidates:
            sequence = candidates[candidate]
            temp_fasta = '{0}/{1}.fa'.format(outdir, candidate)
            with open(temp_fasta, 'w') as tempfile:
                tempfile.write('>{0}\n{1}\n'.format(candidate, sequence))
            blast_output = '{0}/blast_{1}.out'.format(outdir, candidate)
            print('# Starting reverse blast for {}'.format(candidate))
            # blast_search will write results to the temp_fasta file
            blast_command = (
                'blastn -task blastn -db {0} -query {1} -max_target_seqs 10 -dust {4} '
                '-out {2} -num_threads {3} -outfmt "6 qseqid sseqid pident '
                'length mismatch gapopen qstart qend sstart send evalue bitscore sseq"'
                    .format(refblast, temp_fasta, blast_output, cpu, dust)
            )
            sp.run(blast_command, shell=True)
            print('# finished blast')

            # BlastParser will read the temp_fasta file
            bp = BlastParser(mirna, blast_output, msl)
            # the parse_blast_output function will return True if the candidate is accepted
            if bp.evaluate_besthit():
                print('Found best hit')
                reblast_hits[candidate] = sequence
            elif checkCoorthref:
                print('Best hit differs from reference sequence! Doing further checks\n')
                # coorth_out = '{0}/{1}_coorth'.format(outdir, candidate)
                if bp.check_coortholog_ref(sequence, outdir):
                    reblast_hits[candidate] = sequence
            else:
                print('Best hit does not overlap with miRNA location')
            if cleanup:
                os.remove(temp_fasta)
                os.remove(blast_output)

        # Write output file if at least one candidate got accepted.
        if reblast_hits:
            nr_accepted = len(reblast_hits)
            if nr_accepted == 1:
                print('# ncOrtho found 1 verified ortholog.\n')
                out_dict = reblast_hits
            else:
                print(
                    '# ncOrtho found {} potential co-orthologs.\n'
                    .format(nr_accepted)
                )
                print('# Evaluating distance between candidates to verify co-orthologs')
                rbp = ReBlastParser(mirna, reblast_hits)
                out_dict = rbp.verify_coorthologs(outdir)
                if len(out_dict) == 1:
                    print('ncOrtho found 1 verified ortholog')
                else:
                    print(
                        '# ncOrtho found {} verified co-orthologs.\n'
                            .format(len(out_dict))
                    )

            print('# Writing output of accepted candidates.\n')
            outpath = '{0}/{1}_orthologs.fa'.format(outdir, mirna_id)
            write_output(out_dict, outpath, cm_results)
            print('# Finished writing output.\n')
        else:
            print(
                '# None of the candidates for {} could be verified.\n'
                .format(mirna_id)
            )
        print('# Finished ortholog search for {}.'.format(mirna_id))
    print('### ncOrtho is finished!')


if __name__ == "__main__":
    main()
